

#import matplotlib
#import matplotlib.pyplot as plt
#from matplotlib.backends.backend_agg import FigureCanvasAgg


#import os
import ctypes
import numpy as np
import time

from multiprocessing import Process, Pipe, Array, Value

def tonumpyarray(shared_mem_array):
    return np.frombuffer(shared_mem_array.get_obj(), dtype=np.uint8)

def plot_receiver(connection, shared_mem_for_plot, shared_mem_trainId):
    '''
    this function makes the plot for the single shot contrast'''
    print('Receiver: Running', flush=True)
    # consume work

    
    while True:
        # get a unit of work
        item = connection.recv()

        if item is None:
            break
        
        output = item['fun'](*item['arg']) 
        npArray = tonumpyarray(shared_mem_for_plot)
        npArray[:] = output.reshape(-1)
        shared_mem_trainId.value = item['train_id']


        
    # all done
    print('Receiver: Done', flush=True)


class OnlineImagePlotter():

    def __init__(self, figsize = (15,10)):
        self.figsize = figsize
        arraysize_bytes = (self.figsize[0]*100)*(self.figsize[1]*100)*4

        self.plot_memory = Array(ctypes.c_ubyte, [0]*arraysize_bytes)
        self.train_id =  Value(ctypes.c_ulonglong, 0)
        self.prev_train_id = -1

        #pipe
        receiver, sender = Pipe()
        self.receiver = receiver
        self.sender = sender

        self.process = Process(target=plot_receiver, args=(self.receiver, self.plot_memory, self.train_id), daemon=True)
        self.process.start()

    def get_image_buffer_if_ready_and_make_new_image(self, train_id, fun, *args):
        '''
        return None if image is not ready
        return np array of shape (fihsize[1]*100, figsize[1]*100, 4) otherwise
        which then can be used in the output hash
        -> output_hash[' hash name '] = ImageData(np array from this function)
        '''

        current_image = None

        def is_new_image_ready():
            return self.prev_train_id != self.train_id.value

        if is_new_image_ready():
            
            # get the previously computed image

            def get_plot_as_np_array(shared_mem_array):
                np_array = tonumpyarray(shared_mem_array)
                return np.reshape(np_array, (self.figsize[1]*100, self.figsize[0]*100, 4))
            
            current_image = get_plot_as_np_array(self.plot_memory)

            self.prev_train_id = self.train_id.value

            # ask for next image to be made
            package = dict()
            package['train_id'] = train_id
            package['fun'] = fun
            package['arg'] = args
            self.sender.send(package)
        else:
            pass
        return current_image


class DataToPlotToHash():

    def __init__(self):

        self.hash_keys_and_plotter = dict()

    def is_hash_key_registered(self, hash_key):
        if hash_key not in self.hash_keys_and_plotter:
            print(f"key {hash_key} does not exist, call 'register' function in initialization() part of this karabo device!")
            return hash
        
    def register(self, hash_key, Plotter):
        if hash_key not in self.hash_keys_and_plotter:
            print(f"key {hash_key} does not exist, init plotter")
            self.hash_keys_and_plotter[hash_key] = { 
                "plotter" : Plotter(),
                "time_to_make_plot" : time.time()}
        
    
    
    def add_schema(self, hash_key, schema):
        if not self.is_hash_key_registered(hash_key):
            return schema
        
        from karabo.bound import (
            NODE_ELEMENT,
            IMAGEDATA_ELEMENT,
            OUTPUT_CHANNEL,
            STRING_ELEMENT,
            Schema
            )

            
        basenode = hash_key
        (
            NODE_ELEMENT(schema)
            .key(basenode)
            .displayedName(f'{basenode} - agipd module plot')
            .commit(),

            IMAGEDATA_ELEMENT(schema)
            .key(f"{basenode}.plot")
            .displayedName('plot')
            .description('plot')
            .commit(),

            STRING_ELEMENT(schema)
            .key(f"{basenode}.update_interval")
            .displayedName('Update interval')
            .description('Update interval of the plot')
            .assignmentOptional()
            .defaultValue("")
            .commit(),
        )

        return schema
        
    def add_to_hash(self, hash, hash_key, *args):
        
        if hash_key not in self.hash_keys_and_plotter:
            print(f"key {hash_key} does not exist, call get_schema_update function in initialization() part of this karabo device!")
            return hash
        
        # plotter is initialized, give plotter data
        current_img = self.hash_keys_and_plotter[hash_key]['plotter'].plot(*args)

        from karabo.bound import ImageData

        if current_img is not None:
            hash[f'{hash_key}.plot'] = ImageData(current_img)

            now = time.time()
            hash[f'{hash_key}.update_interval'] = f'{(now - self.hash_keys_and_plotter[hash_key]['time_to_make_plot']):.1f}s' 
            self.hash_keys_and_plotter[hash_key]['time_to_make_plot'] = now
        
        return hash
    
